#
# Virtuozzo containers hauler module
#

import logging
import os
import shlex
import subprocess
import util
import signal
import select
import glob

import pycriu.rpc_pb2

import criu_api
import criu_cr
import fs_haul_ploop
import pycriu


vz_global_conf = "/etc/vz/vz.conf"
vz_conf_dir = "/etc/vz/conf/"
vzctl_bin = "vzctl"
cgget_bin = "cgget"
cgexec_bin = "cgexec"


vz_cgroup_mount_map = {
	"cpu,cpuacct": "cpu",
	"cpuset": "cpuset",
	"net_cls": "net_cls",
	"net_cls,net_prio": "net_cls,net_prio",
	"memory": "memory",
	"devices": "devices",
	"blkio": "blkio",
	"freezer": "freezer",
	"perf_event": "perf_event",
	"hugetlb": "hugetlb",
	"name=systemd": "systemd",
	"pids": "pids",
}

cgroup_controllers = [
	"cpuacct,cpu",
	"cpuset",
	"net_cls",
	"net_prio,net_cls",
	"memory",
	"devices",
	"blkio",
	"freezer",
	"perf_event",
	"hugetlb",
	"name=systemd",
	"pids",
	"beancounter",
	"ve",
]


vz_action_scripts = {
	"post-network-lock": ["/usr/libexec/criu/scripts/nfs-ports-allow.sh"],
}


class p_haul_type(object):
	def __init__(self, ctid):
		self._ctid = ctid
		self._ct_priv = ""
		self._ct_root = ""
		#
		# This list would contain (v_in, v_out, v_br) tuples where
		# v_in is the name of veth device in CT
		# v_out is its peer on the host
		# v_bridge is the bridge to which thie veth is attached
		#
		self._veths = []
		self.__verbose = criu_api.def_verb
		self._args_path = ''
		self._resume_wait_pipe = []
		self._vzctl_proc = 0
		self._has_nfs = False
		self._has_dm_disk = False

	def __load_ct_config(self, path):
		logging.info("Loading config file from %s", path)

		# Read container config
		with open(self.__ct_config_path(path)) as ifd:
			config = _parse_vz_config(ifd.read())

		# Read global config
		with open(vz_global_conf) as ifd:
			global_config = _parse_vz_config(ifd.read())

		# Extract veth pairs, later we will equip restore request with this
		# data and will use it while (un)locking the network
		if "NETIF" in config:
			v_in, v_out, v_bridge = None, None, None
			for parm in config["NETIF"].split(","):
				pa = parm.split("=")
				if pa[0] == "ifname":
					v_in = pa[1]
				elif pa[0] == "host_ifname":
					v_out = pa[1]
				elif pa[0] == "bridge":
					v_bridge = pa[1]
			if v_in and v_out:
				logging.info("\tCollect %s -> %s (%s) veth",
							v_in, v_out, v_bridge)
				self._veths.append(util.net_dev(v_in, v_out, v_bridge))

		# Extract private path from config
		if "VE_PRIVATE" in config:
			self._ct_priv = _expand_veid_var(config["VE_PRIVATE"], self._ctid)
		else:
			self._ct_priv = _expand_veid_var(
				global_config["VE_PRIVATE"], self._ctid)

		# Extract root path from config
		if "VE_ROOT" in config:
			self._ct_root = _expand_veid_var(config["VE_ROOT"], self._ctid)
		else:
			self._ct_root = _expand_veid_var(
				global_config["VE_ROOT"], self._ctid)

	def __load_ct_config_dst(self, path):
		if not os.path.isfile(self.__ct_config_path(path)):
			raise Exception("CT config missing on destination")
		self.__load_ct_config(path)

	def __ct_config_path(self, conf_dir):
		return os.path.join(conf_dir, "{0}.conf".format(self._ctid))

	def __cg_set_veid(self):
		"""Initialize veid in ve.veid for ve cgroup"""

		veid_path = "/sys/fs/cgroup/ve/{0}/ve.veid".format(self._ctid)
		with open(veid_path, "w") as f:
			if self._ctid.isdigit():
				veid = self._ctid
			else:
				veid = str(int(self._ctid.partition("-")[0], 16))
			f.write(veid)

	def __has_nfs_mount(self, img):
		"""Check if NFS mount is present in CRIU images"""
		for image in glob.glob(img.image_dir() + "/mountpoints-*.img"):
			with open(image, "r") as f:
				try:
					mp_img = pycriu.images.load(f)
				except:
					raise Exception("Unable to read mountpoints CRIU image")
				for entry in mp_img["entries"]:
					if entry["fstype"] == 21 or entry["fstype"] == 22:
						logging.info("NFS mount point detected: ({0} : {1})".format(entry["source"], entry["mountpoint"]))
						return True
		return False

	def init_src(self):
		self._fs_mounted = True
		self._bridged = True
		self.__load_ct_config(vz_conf_dir)

	def init_dst(self):
		self._fs_mounted = False
		self._bridged = False
		self.__load_ct_config_dst(vz_conf_dir)

	def set_options(self, opts):
		self.__verbose = opts["verbose"]
		self.__secondary_disks = opts.get("vz_secondary_disks")

	def adjust_criu_req(self, req):
		"""Add module-specific options to criu request"""
		# Specify dump specific options
		if req.type == pycriu.rpc.DUMP:

			# Specify root fs
			req.opts.root = self._ct_root

			# Restore cgroups configuration
			req.opts.manage_cgroups = True

			# Setup mapping for external Virtuozzo specific cgroup mounts
			self.__fill_vz_cgroup_ext_mount_map(req)

			# Specify secondary ploop disks as external
			disks = self.__parse_secondary_disks_arg(self.__secondary_disks)
			for uuid, major, minor in disks:
				req.opts.external.append(
					"dev[{0}/{1}]:{2}".format(major, minor, uuid))

			# Increase ghost-limit up to 4Gb
			req.opts.ghost_limit = 4 << 30 - 1

		# Specify both predump and dump specific options
		if req.type == pycriu.rpc.PRE_DUMP or req.type == pycriu.rpc.DUMP:

			# Specify freezer cgroup
			req.opts.freeze_cgroup = \
				"/sys/fs/cgroup/freezer/machine.slice/{0}/".format(self._ctid)

			# Setup cgroup dump controllers
			for cgname in cgroup_controllers:
				req.opts.cgroup_dump_controller.append(cgname)

			# Increase timeout up to 180 seconds
			req.opts.timeout = 180

	def __fill_vz_cgroup_ext_mount_map(self, req):
		with open("/proc/self/mountinfo") as x:
			mountinfo = x.readlines()
		# Walk cgroups maps
		for cgkey in vz_cgroup_mount_map:
			# Iterate through lines in mountinfo until we find a match
			for buf in mountinfo:
				cg_matches = True
				mi_line = buf.split()
				if mi_line[8] != "cgroup":
					continue
				# If key consists of two cgroups (e.g."cpu,cpuacct"), search each
				for cgname in cgkey.split(","):
					if cgname not in mi_line[10].split(","):
						cg_matches = False
						break
				if cg_matches:
					req.opts.ext_mnt.add(key = mi_line[4],
							val = vz_cgroup_mount_map[cgkey])
					break

	def root_task_pid(self):
		path = "/var/run/ve/{0}.init.pid".format(self._ctid)
		with open(path) as pidfile:
			pid = pidfile.read()
			return int(pid)

	def get_meta_images(self, path):
		return []

	def put_meta_images(self, path):
		pass

	def ve_sync(self, timeout, fail_on_timeout):
		"""Perform sync in a container context"""
		logging.info("  Sync VE filesystems")
		signal.signal(signal.SIGALRM, util.alarm_handler)
		signal.alarm(timeout)
		try:
			proc = subprocess.Popen(
				[cgexec_bin, "-g", "ve,beancounter:{0}".format(self._ctid), "/bin/sync"],
				stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
			proc_output = proc.communicate()[0]
			signal.alarm(0)
		except util.Alarm:
			if fail_on_timeout:
				raise Exception("VE filesystems sync timed out ({0} seconds)".format(timeout))
			else:
				logging.warning("VE filesystems sync timed out (non-blocking): ({0} seconds)".format(timeout))
				proc.kill()
		if proc.returncode:
			raise Exception("VE filesystems sync failed ({0}):{1}".format(proc.returncode, proc_output))

	def final_dump(self, pid, img, ccon, fs):
		"""Perform Virtuozzo-specific final dump"""
		self.__pre_final_dump(img)
		criu_cr.criu_dump(self, pid, img, ccon, fs)
		self.__post_final_dump(img)

	def __pre_final_dump(self, img):
		"""Sync ve and create extra images before final dump"""
		if self._has_nfs or self._has_dm_disk:
			self.ve_sync(150, False)

		extra_images = (
			("vz_clock_bootbased.img", "ve.clock_bootbased"),
			("vz_clock_monotonic.img", "ve.clock_monotonic"),
			("vz_iptables_mask.img", "ve.iptables_mask"),
			("vz_os_release.img", "ve.os_release"),
			("vz_features.img", "ve.features"),
			("vz_aio_max_nr.img", "ve.aio_max_nr"))
		for image_name, var_name in extra_images:
			self.__create_cgget_extra_image(img, image_name, var_name)

	def __post_final_dump(self, img):
		"""Sync ve and create extra images after final dump"""
		if self._has_nfs or self._has_dm_disk:
			self.ve_sync(5, True)

		extra_images = (
			("vz_core_pattern.img", ["cat", "/proc/sys/kernel/core_pattern"]),)
		for image_name, exec_args in extra_images:
			self.__create_cgexec_extra_image(img, image_name, exec_args)

	def __create_cgget_extra_image(self, img, image_name, var_name):
		"""Create extra image using cgget output"""
		proc = subprocess.Popen(
			[cgget_bin, "-n", "-v", "-r", var_name, self._ctid],
			stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
		image_data = proc.communicate()[0]
		if proc.returncode == 0:
			self.__create_extra_image(img, image_name, image_data)
		else:
			logging.warning("cgget failed to create %s", image_name)

	def __create_cgexec_extra_image(self, img, image_name, exec_args):
		"""Create extra image using cgexec output"""
		proc = subprocess.Popen(
			[cgexec_bin, "-g", "ve:{0}".format(self._ctid)] + exec_args,
			stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
		image_data = proc.communicate()[0]
		if proc.returncode == 0:
			self.__create_extra_image(img, image_name, image_data)
		else:
			logging.warning("cgexec failed to create %s", image_name)

	def __create_extra_image(self, img, image_name, image_data):
		image_path = os.path.join(img.image_dir(), image_name)
		with open(image_path, "w") as f:
			f.write(image_data)

	def _finalize_restore(self):
		"""Finalize restore"""
		if self._has_nfs:
			logging.info("\tContinue resume process")
			os.write(self._resume_wait_pipe[1], "\0\0\0\0")

		proc_output = self._vzctl_proc.communicate()[0]
		util.log_subprocess_output(proc_output)
		try:
			if self._vzctl_proc.returncode != 0:
				raise Exception("Restore failed ({0}):{1}".format(self._vzctl_proc.returncode, self._vzctl_proc.communicate()[0]))
		finally:
			# Remove restore extra arguments
			self.__remove_restore_extra_args(self._args_path)

	def final_restore(self, img, connection):
		"""Perform Virtuozzo-specific final restore"""
		# Setup restore extra arguments
		self._args_path = os.path.join(img.image_dir(), "restore-extra-args")
		self.__setup_restore_extra_args(self._args_path, img, connection)

		# Check if container has NFS mounts
		self._has_nfs = self.__has_nfs_mount(img)

		# Prepare wait pipes for action scripts
		ph_env = os.environ.copy()
		if self._has_nfs:
			ph_pipe_wait, vz_pipe_write = os.pipe()
			self._resume_wait_pipe = os.pipe()
			ph_env["CRIU_ACTION_POST_RESUME_READ_FD"] = str(vz_pipe_write)
			ph_env["CRIU_ACTION_POST_RESUME_WRITE_FD"] = str(self._resume_wait_pipe[0])

		# Run vzctl restore
		logging.info("Starting vzctl restore")
		self._vzctl_proc = subprocess.Popen([vzctl_bin, "--skipowner", "--skiplock",
			"restore", self._ctid, "--skip_arpdetect",
			"--dumpfile", img.image_dir()], env=ph_env,
			stdout=subprocess.PIPE,
			stderr=subprocess.STDOUT)

		if self._has_nfs:
			# Close write end of the pipe to make sure pipe dies if child exits
			# so we can poll just the pipe, instead of checking both child and pipe
			os.close(vz_pipe_write)
			os.close(self._resume_wait_pipe[0])
			select.select([ph_pipe_wait], [], [])

			if self._vzctl_proc.poll():
				self.__remove_restore_extra_args(self._args_path)
				raise Exception("Restore failed ({0}):{1}".format(self._vzctl_proc.returncode, self._vzctl_proc.communicate()[0]))
			os.read(ph_pipe_wait, 4)
			logging.info("\tReached post-restore stage, kill source container early for NFS scenario")
		else:
			self._finalize_restore()

	def __setup_restore_extra_args(self, path, img, connection):
		"""Create temporary file with extra arguments for criu restore"""
		extra_args = [
			"VE_WORK_DIR={0}\n".format(img.work_dir()),
			"VE_RESTORE_LOG_PATH={0}\n".format(
				connection.get_log_name(pycriu.rpc.RESTORE)),
			"VE_CRIU_LOGLEVEL={0}\n".format(self.__verbose)]
		with open(path, "w") as f:
			f.writelines(extra_args)

	def __remove_restore_extra_args(self, path):
		"""Remove temporary file with extra arguments for criu restore"""
		if os.path.isfile(path):
			os.remove(path)

	def prepare_ct(self, pid):
		"""Create cgroup hierarchy and put root task into it."""
		self.__cg_set_veid()

	def mount(self):
		logging.info("Mounting CT root to %s", self._ct_root)
		logging.info("Running vzctl mount")
		proc = subprocess.Popen(
			[vzctl_bin, "--skipowner", "--skiplock", "mount", self._ctid],
			stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
		proc_output = proc.communicate()[0]
		util.log_subprocess_output(proc_output)
		self._fs_mounted = True
		return self._ct_root

	def umount(self):
		if self._fs_mounted:
			logging.info("Umounting CT root")
			logging.info("Running vzctl umount")
			proc = subprocess.Popen(
				[vzctl_bin, "--skipowner", "--skiplock", "stop", self._ctid, "--force"],
				stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
			proc_output = proc.communicate()[0]
			util.log_subprocess_output(proc_output)
			self._fs_mounted = False

	def migration_complete(self, fs, target_host):
		self.umount()
		fs.cleanup_shared_ploops()
		target_host.migration_complete(fs.prepare_src_data({}))

	def migration_fail(self, fs):
		fs.restore_shared_ploops()

	def target_cleanup(self, src_data):
		# Finalize restore for NFS case
		if self._has_nfs:
			self._finalize_restore()
		# Cleanup shared ploops
		if "shareds" in src_data:
			for ploop in src_data["shareds"]:
				fs_haul_ploop.merge_ploop_snapshot(
					ploop["ddxml"], ploop["guid"])

	def start(self):
		logging.info("Starting CT")
		logging.info("Running vzctl start")
		proc = subprocess.Popen(
			[vzctl_bin, "--skipowner", "--skiplock", "start", self._ctid],
			stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
		proc_output = proc.communicate()[0]
		util.log_subprocess_output(proc_output)
		if proc.returncode != 0:
			raise Exception("Start failed ({0})".format(proc.returncode))
		self._fs_mounted = True

	def stop(self, umount):
		logging.info("Stopping CT")
		logging.info("Running vzctl stop")
		args = [vzctl_bin, "--skipowner", "--skiplock", "stop", self._ctid]
		if not umount:
			args.append("--skip-umount")
		proc = subprocess.Popen(
			args, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
		proc_output = proc.communicate()[0]
		util.log_subprocess_output(proc_output)
		self._fs_mounted = not umount

	def get_fs(self, fdfs=None):
		deltas = self.__parse_fdfs_arg(fdfs)
		return fs_haul_ploop.p_haul_fs(deltas, self._ct_priv)

	def get_fs_receiver(self, fdfs=None):
		deltas = self.__parse_fdfs_arg(fdfs)
		return fs_haul_ploop.p_haul_fs_receiver(deltas)

	def restored(self, pid):
		pass

	def net_lock(self):
		pass

	def net_unlock(self):
		pass

	def run_action_scripts(self, stage):
		if stage not in vz_action_scripts:
			return
		script_env = os.environ.copy()
		script_env["CRTOOLS_INIT_PID"] = str(self.root_task_pid())
		script_env["CRTOOLS_SCRIPT_ACTION"] = stage
		for script_bin in vz_action_scripts[stage]:
			child = subprocess.Popen(script_bin, env=script_env)
			child.wait()
			logging.info("\t\t\tAction script %s finished with exit code %d" % (script_bin, child.returncode))

	def can_migrate_tcp(self):
		return True

	def can_pre_dump(self):
		return True

	def dump_need_page_server(self):
		return True

	def __parse_fdfs_arg(self, fdfs):
		"""Parse string containing list of ploop deltas with socket fds

		String contain list of active ploop deltas with corresponding socket
		file descriptors in format %delta_path1%:%socket_fd1%[,...]. Parse it
		and return list of tuples.
		"""
		deltas = []
		for path, fd in util.parse_fdfs(fdfs):
			delta_path = fs_haul_ploop.get_delta_abspath(path, self._ct_priv)
			deltas.append((delta_path, int(fd)))

		return deltas

	def __parse_secondary_disks_arg(self, disks_arg):
		"""Parse string containing list of secondary disks in specific format

		String contain list of secondary ploop disks in format
		%uuid%:%major%:%minor%[,...]. Consider all additional disks of
		container (second, third and so on) are secondary. Parse it and return
		list of tuples.
		"""

		DISKS_SEPARATOR = ","
		TUPLE_SEPARATOR = ":"

		if not disks_arg:
			return []

		disks = []
		for disk in disks_arg.split(DISKS_SEPARATOR):
			uuid, major, minor = disk.split(TUPLE_SEPARATOR)
			disks.append((uuid, int(major), int(minor)))
			if int(major) == 253:
				self._has_dm_disk = True

		return disks


def add_hauler_args(parser):
	"""Add Virtuozzo specific command line arguments"""
	parser.add_argument("--vz-shared-disks",
						help="List of shared storage disks")
	parser.add_argument("--vz-secondary-disks",
						help="List of secondary ploop disks")


def _parse_vz_config(body):
	"""Parse shell-like virtuozzo config file"""

	config_values = dict()
	for token in shlex.split(body, comments=True):
		name, sep, value = token.partition("=")
		config_values[name] = value
	return config_values


def _expand_veid_var(value, ctid):
	"""Replace shell-like VEID variable with actual container id"""
	return value.replace("$VEID", ctid).replace("${VEID}", ctid)
